#include "COP_VU.h"
#include "VUShared.h"
#include "MIPS.h"

CCOP_VU			g_COPVU(MIPS_REGSIZE_64);

uint8			CCOP_VU::m_nFT		= 0;
uint8			CCOP_VU::m_nFS		= 0;
uint8			CCOP_VU::m_nFD		= 0;

uint8			CCOP_VU::m_nDest	= 0;
uint8			CCOP_VU::m_nFTF		= 0;
uint8			CCOP_VU::m_nFSF		= 0;
uint8			CCOP_VU::m_nBc		= 0;

CCOP_VU::CCOP_VU(MIPS_REGSIZE nRegSize) :
CMIPSCoprocessor(nRegSize)
{
	SetupReflectionTables();
}

void CCOP_VU::CompileInstruction(uint32 nAddress, CCacheBlock* pBlock, CMIPS* pCtx, bool nParent)
{
	if(nParent)
	{
		SetupQuickVariables(nAddress, pBlock, pCtx);
	}

	m_nDest			= (uint8)((m_nOpcode >> 21) & 0x0F);
	
	m_nFSF			= ((m_nDest >> 0) & 0x03);
	m_nFTF			= ((m_nDest >> 2) & 0x03);

	m_nFT			= (uint8)((m_nOpcode >> 16) & 0x1F);
	m_nFS			= (uint8)((m_nOpcode >> 11) & 0x1F);
	m_nFD			= (uint8)((m_nOpcode >>  6) & 0x1F);

	m_nBc			= (uint8)((m_nOpcode >>  0) & 0x03);

	switch((m_nOpcode >> 26) & 0x3F)
	{
	case 0x12:
		//COP2
		m_pOpCop2[(m_nOpcode >> 21) & 0x1F]();
		break;
	case 0x36:
		//LQC2
		LQC2();
		break;
	case 0x3E:
		//SQC2
		SQC2();
		break;
	default:
		Illegal();
		break;
	}
}

//////////////////////////////////////////////////
//General Instructions
//////////////////////////////////////////////////

//36
void CCOP_VU::LQC2()
{
	ComputeMemAccessAddr();

	//Load the word
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV0);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV1);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 1, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV2);
	m_pB->AddImm(4);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV3);
}

//3E
void CCOP_VU::SQC2()
{
	ComputeMemAccessAddr();

	//Write the words
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV0);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV1);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV2);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 2, false);
	m_pB->AddImm(4);

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2[m_nFT].nV3);
	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 3, false);	
}

//////////////////////////////////////////////////
//COP2 Instructions
//////////////////////////////////////////////////

//02
void CCOP_VU::CFC2()
{
	m_pB->PushImm(0);
	SignExtendTop32(m_nFT);
	m_pB->PullAddr(&m_pCtx->m_State.nGPR[m_nFT].nV[0]);
}

//06
void CCOP_VU::CTC2()
{

}

//10-1F
void CCOP_VU::V()
{
	m_pOpVector[(m_nOpcode & 0x3F)]();
}

//////////////////////////////////////////////////
//Vector Instructions
//////////////////////////////////////////////////

//01
//02
void CCOP_VU::VADDbc()
{
	VUShared::ADDbc(m_pB, m_pCtx, m_nDest, m_nFD, m_nFS, m_nFT, m_nBc);
}

//0B
void CCOP_VU::VMADDbc()
{
	VUShared::MADDbc(m_pB, m_pCtx, m_nDest, m_nFD, m_nFS, m_nFT, m_nBc);
}

//1C
void CCOP_VU::VMULq()
{
	VUShared::MULq(m_pB, m_pCtx, m_nDest, m_nFD, m_nFS);
}

//28
void CCOP_VU::VADD()
{
	VUShared::ADD(m_pB, m_pCtx, m_nDest, m_nFD, m_nFS, m_nFT);
}

//2A
void CCOP_VU::VMUL()
{
	VUShared::MUL(m_pB, m_pCtx, m_nDest, m_nFD, m_nFS, m_nFT);
}

//2E
void CCOP_VU::VOPMSUB()
{
	VUShared::OPMSUB(m_pB, m_pCtx, m_nFD, m_nFS, m_nFT);
}

//3C
void CCOP_VU::VX0()
{
	m_pOpVx0[(m_nOpcode >> 6) & 0x1F]();
}

//3D
void CCOP_VU::VX1()
{
	m_pOpVx1[(m_nOpcode >> 6) & 0x1F]();
}

//3E
void CCOP_VU::VX2()
{
	m_pOpVx2[(m_nOpcode >> 6) & 0x1F]();
}

//3F
void CCOP_VU::VX3()
{
	m_pOpVx3[(m_nOpcode >> 6) & 0x1F]();
}

//////////////////////////////////////////////////
//Vx Common Instructions
//////////////////////////////////////////////////

//
void CCOP_VU::VMULAbc()
{
	VUShared::MULAbc(m_pB, m_pCtx, m_nDest, m_nFS, m_nFT, m_nBc);
}

//
void CCOP_VU::VMADDAbc()
{
	VUShared::MADDAbc(m_pB, m_pCtx, m_nDest, m_nFS, m_nFT, m_nBc);
}

//////////////////////////////////////////////////
//V2 Instructions
//////////////////////////////////////////////////

//0B
void CCOP_VU::VOPMULA()
{
	VUShared::OPMULA(m_pB, m_pCtx, m_nFS, m_nFT);
}

//0E
void CCOP_VU::VRSQRT()
{
	VUShared::RSQRT(m_pB, m_pCtx, m_nFS, m_nFSF, m_nFT, m_nFTF);
}

//////////////////////////////////////////////////
//V3 Instructions
//////////////////////////////////////////////////

//0E
void CCOP_VU::VWAITQ()
{
	//Nothing to do
}

//////////////////////////////////////////////////
//Opcode Tables
//////////////////////////////////////////////////

void (*CCOP_VU::m_pOpCop2[0x20])() =
{
	//0x00
	Illegal,		Illegal,		CFC2,			Illegal,		Illegal,		Illegal,		CTC2,			Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	V,				V,				V,				V,				V,				V,				V,				V,
	//0x18
	V,				V,				V,				V,				V,				V,				V,				V,
};

void (*CCOP_VU::m_pOpVector[0x40])() =
{
	//0x00
	Illegal,		VADDbc,			VADDbc,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		VMADDbc,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		VMULq,			Illegal,		Illegal,		Illegal,
	//0x20
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x28
	VADD,			Illegal,		VMUL,			Illegal,		Illegal,		Illegal,		VOPMSUB,		Illegal,
	//0x30
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x38
	Illegal,		Illegal,		Illegal,		Illegal,		VX0,			VX1,			VX2,			VX3,
};

void (*CCOP_VU::m_pOpVx0[0x20])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		VMULAbc,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CCOP_VU::m_pOpVx1[0x20])() =
{
	//0x00
	Illegal,		Illegal,		VMADDAbc,		Illegal,		Illegal,		Illegal,		VMULAbc,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CCOP_VU::m_pOpVx2[0x20])() =
{
	//0x00
	Illegal,		Illegal,		VMADDAbc,		Illegal,		Illegal,		Illegal,		VMULAbc,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		VOPMULA,		Illegal,		Illegal,		VRSQRT,			Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CCOP_VU::m_pOpVx3[0x20])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		VWAITQ,			Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};
