#ifndef _CODEGEN_H_
#define _CODEGEN_H_

#include "CacheBlock.h"
#include "ArrayStack.h"

class CCodeGen
{
public:
	enum CONDITION
	{
		CONDITION_EQ,
		CONDITION_NE,
		CONDITION_BL,
		CONDITION_BE,
	};

	static void					Begin(CCacheBlock*);
	static void					End();

	static void					BeginJump(bool);
	static void					EndJump();

	static void					PushVar(uint32*);
	static void					PushCst(uint32);

	static void					PullVar(uint32*);

	static void					Add();
	static void					Add64();
	static void					And();
	static void					Cmp(CONDITION);
	static void					Cmp64(CONDITION);
	static void					Or();
	static void					SeX();

private:
	enum MAX_STACK
	{
		MAX_STACK = 0x100,
	};

	enum MAX_REGISTER
	{
		MAX_REGISTER = 6,
	};

	enum SYMBOLS
	{
		VARIABLE = 0x8000,
		REGISTER,
		CONSTANT,
	};

	static unsigned int			AllocateRegister();
	static void					FreeRegister(unsigned int);
	static unsigned int			GetMinimumConstantSize(uint32);
	static bool					RegisterHasNextUse(unsigned int);

	static void					Cmp64Eq();
	static void					Cmp64Bl();

	static bool					m_nBlockStarted;

	static CArrayStack<uint32>	m_Shadow;
	static bool					m_nRegisterAllocated[MAX_REGISTER];
	static unsigned int			m_nRegisterLookup[MAX_REGISTER];
	static CCacheBlock*			m_pBlock;

	static unsigned int			m_nJumpPosition;
};

#endif
