#include <stdio.h>
#include "GIF.h"
#include "uint128.h"
#include "PS2VM.h"

uint16	CGIF::m_nLoops		= 0;
uint8	CGIF::m_nCmd		= 0;
uint8	CGIF::m_nRegs		= 0;
uint64	CGIF::m_nRegList	= 0;
bool	CGIF::m_nEOP		= false;

void CGIF::Reset()
{
	m_nLoops = 0;
	m_nCmd = 0;
	m_nEOP = false;
}

uint32 CGIF::ProcessPacked(uint8* pMemory, uint32 nAddress, uint32 nEnd)
{
	uint32 nRegDesc, j, nStart;
	uint64 nTemp;
	uint128 nPacket;

	nStart = nAddress;

	while(m_nLoops != 0)
	{
		for(j = 0; j < m_nRegs; j++)
		{
			nRegDesc = (uint32)((m_nRegList >> (j * 4)) & 0x0F);

			nPacket.nV[0] = *(uint32*)&pMemory[nAddress + 0x00];
			nPacket.nV[1] = *(uint32*)&pMemory[nAddress + 0x04];
			nPacket.nV[2] = *(uint32*)&pMemory[nAddress + 0x08];
			nPacket.nV[3] = *(uint32*)&pMemory[nAddress + 0x0C];
			nAddress += 0x10;

			switch(nRegDesc)
			{
			case 0x01:
				//RGBA
				nTemp  = (nPacket.nV[0] & 0xFF);
				nTemp |= (nPacket.nV[1] & 0xFF) << 8;
				nTemp |= (nPacket.nV[2] & 0xFF) << 16;
				nTemp |= (nPacket.nV[3] & 0xFF) << 24;
				CPS2VM::m_pGS->WriteRegister(GS_REG_RGBAQ, nTemp);
				break;
			case 0x03:
				//UV
				nTemp  = (nPacket.nV[0] & 0x7FFF);
				nTemp |= (nPacket.nV[1] & 0x7FFF) << 16;
				CPS2VM::m_pGS->WriteRegister(GS_REG_UV, nTemp);
				break;
			case 0x04:
				//XYZF2
				nTemp  = (nPacket.nV[0] & 0xFFFF);
				nTemp |= (nPacket.nV[1] & 0xFFFF) << 16;
				nTemp |= (uint64)(nPacket.nV[2] & 0xFFFFFF) << 32;
				CPS2VM::m_pGS->WriteRegister(GS_REG_XYZ2, nTemp);
				break;
			case 0x0E:
				//A + D
				CPS2VM::m_pGS->WriteRegister((uint8)nPacket.nD1, nPacket.nD0);
				break;
			default:
				assert(0);
				break;
			}
		}

		m_nLoops--;
	}

	return nAddress - nStart;
}

uint32 CGIF::ProcessImage(uint8* pMemory, uint32 nAddress, uint32 nEnd)
{
	uint16 nTotalLoops;
	
	nTotalLoops = (uint16)((nEnd - nAddress) / 0x10);
	nTotalLoops = min(nTotalLoops, m_nLoops);

	CPS2VM::m_pGS->FeedImageData(pMemory + nAddress, nTotalLoops * 0x10);

	m_nLoops -= nTotalLoops;

	return (nTotalLoops * 0x10);
}

uint32 CGIF::ProcessPacket(uint8* pMemory, uint32 nAddress, uint32 nEnd)
{
	uint128 nPacket;
	uint32 nStart;

	if(CPS2VM::m_Logging.GetGSLoggingStatus())
	{
		printf("GIF: Processed GIF packet at 0x%0.8X, PC: 0x%0.8X.\r\n", nAddress, CPS2VM::m_EE.m_State.nPC);
	}

	nStart = nAddress;
	while(nAddress < nEnd)
	{
		if(m_nLoops == 0)
		{
			if(m_nEOP)
			{
				m_nEOP = false;
				break;
			}

			//We need to update the registers
			nPacket.nV[0] = *(uint32*)&pMemory[nAddress + 0x00];
			nPacket.nV[1] = *(uint32*)&pMemory[nAddress + 0x04];
			nPacket.nV[2] = *(uint32*)&pMemory[nAddress + 0x08];
			nPacket.nV[3] = *(uint32*)&pMemory[nAddress + 0x0C];
			nAddress += 0x10;

			if(nPacket.nV1 & 0x4000)
			{
				CPS2VM::m_pGS->WriteRegister(GS_REG_PRIM, (uint16)(nPacket.nV1 >> 15) & 0x3FF);
			}

			m_nLoops	= (uint16)((nPacket.nV0 >>  0) & 0x7FFF);
			m_nCmd		= ( uint8)((nPacket.nV1 >> 26) & 0x0003);
			m_nRegs		= ( uint8)((nPacket.nV1 >> 28) & 0x000F);
			m_nRegList	= nPacket.nD1;
			m_nEOP		= (nPacket.nV0 & 0x8000) != 0;

			if(m_nRegs == 0) m_nRegs = 0x10;
			continue;
		}
		switch(m_nCmd)
		{
		case 0x00:
			nAddress += ProcessPacked(pMemory, nAddress, nEnd);
			break;
		case 0x02:
		case 0x03:
			nAddress += ProcessImage(pMemory, nAddress, nEnd);
			break;
		}
	}

	if(m_nLoops == 0)
	{
		if(m_nEOP)
		{
			m_nEOP = false;
		}
	}

	return nAddress - nStart;
}

uint32 CGIF::ReceiveDMA(uint32 nAddress, uint32 nQWC)
{
	uint32 nEnd, nSize;
	uint8* pMemory;

	if(nAddress & 0x80000000)
	{
		pMemory = CPS2VM::m_pSPR;
		nAddress &= CPS2VM::SPRSIZE - 1;
	}
	else
	{
		pMemory = CPS2VM::m_pRAM;
	}
	
	nSize = nQWC * 0x10;
	nEnd = nAddress + nSize;

	while(nAddress < nEnd)
	{
		nAddress += ProcessPacket(pMemory, nAddress, nEnd);
	}

	return nQWC;
}
