#include <stdio.h>
#include <assert.h>
#include <math.h>
#include "GSH_OpenGL.h"
#include "PS2VM.h"
#include "PtrMacro.h"
#include "RendererSettingsWnd.h"
#include "Config.h"

#ifndef GL_UNSIGNED_SHORT_1_5_5_5_REV
#define GL_UNSIGNED_SHORT_1_5_5_5_REV	(0x8366)
#endif

#ifndef GL_CONSTANT_ALPHA_EXT
#define GL_CONSTANT_ALPHA_EXT			(0x8003)
#endif

#ifndef GL_FUNC_ADD_EXT
#define GL_FUNC_ADD_EXT					(0x8006)
#endif

#ifndef GL_FUNC_REVERSE_SUBTRACT_EXT
#define GL_FUNC_REVERSE_SUBTRACT_EXT	(0x800B)
#endif

#ifndef GL_ONE_MINUS_CONSTANT_ALPHA_EXT
#define GL_ONE_MINUS_CONSTANT_ALPHA_EXT	(0x8004)
#endif

using namespace Framework;

typedef void (APIENTRY * PFNGLBLENDCOLOREXTPROC)(GLclampf, GLclampf, GLclampf, GLclampf);
typedef void (APIENTRY * PFNGLBLENDEQUATIONEXTPROC)(GLenum);

PFNGLCOLORTABLEEXTPROC		glColorTableEXT;
PFNGLBLENDCOLOREXTPROC		glBlendColorEXT;
PFNGLBLENDEQUATIONEXTPROC	glBlendEquationEXT;

PIXELFORMATDESCRIPTOR		CGSH_OpenGL::m_PFD =
{
	sizeof(PIXELFORMATDESCRIPTOR),
	1,
	PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
	PFD_TYPE_RGBA,
	32,
	0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0,
	0,
	32,
	0,
	0,
	PFD_MAIN_PLANE,
	0,
	0, 0, 0
};

CGSH_OpenGL::CGSH_OpenGL(CWindow* pOutputWnd)
{
	m_pOutputWnd = pOutputWnd;
	m_pCvtBuffer = NULL;

	CConfig::GetInstance()->RegisterPreferenceBoolean(PREF_CGSH_OPENGL_LINEASQUADS, false);

	m_nLinesAsQuads = CConfig::GetInstance()->GetPreferenceBoolean(PREF_CGSH_OPENGL_LINEASQUADS);

	InitializeRC();

	m_nVtxCount = 0;
	m_nWidth = -1;
	m_nHeight = -1;
	m_nTexCacheIndex = 0;

	m_nMaxZ = 32768.0;
}

CGSH_OpenGL::~CGSH_OpenGL()
{
	FREEPTR(m_pCvtBuffer);
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(m_hRC);
}

void CGSH_OpenGL::CreateGSHandler(CWindow* pOutputWnd)
{
	CPS2VM::CreateGSHandler(GSHandlerFactory, pOutputWnd);
}

void CGSH_OpenGL::InitializeRC()
{
	unsigned int pf;

	m_hDC = GetDC(m_pOutputWnd->m_hWnd);
	pf = ChoosePixelFormat(m_hDC, &m_PFD);
	SetPixelFormat(m_hDC, pf, &m_PFD);
	m_hRC = wglCreateContext(m_hDC);
	wglMakeCurrent(m_hDC, m_hRC);

	glColorTableEXT = (PFNGLCOLORTABLEEXTPROC)wglGetProcAddress("glColorTable");
	if(glColorTableEXT == NULL)
	{
		glColorTableEXT = (PFNGLCOLORTABLEEXTPROC)wglGetProcAddress("glColorTableEXT");
	}

	glBlendColorEXT	= (PFNGLBLENDCOLOREXTPROC)wglGetProcAddress("glBlendColor");
	if(glBlendColorEXT == NULL)
	{
		glBlendColorEXT	= (PFNGLBLENDCOLOREXTPROC)wglGetProcAddress("glBlendColorEXT");
	}

	glBlendEquationEXT = (PFNGLBLENDEQUATIONEXTPROC)wglGetProcAddress("glBlendEquation");
	if(glBlendEquationEXT == NULL)
	{
		glBlendEquationEXT = (PFNGLBLENDEQUATIONEXTPROC)wglGetProcAddress("glBlendEquationEXT");
	}

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(0.0f);
	glEnable(GL_TEXTURE_2D);

	if(glColorTableEXT == NULL)
	{
		m_pTexUploader_Psm8 = &CGSH_OpenGL::TexUploader_Psm8_Cvt;
	}
	else
	{
		m_pTexUploader_Psm8 = &CGSH_OpenGL::TexUploader_Psm8_Hw;
	}

	VerifyRGBA5551Support();
	if(!m_nIsRGBA5551Supported)
	{
		m_pTexUploader_Psm16 = &CGSH_OpenGL::TexUploader_Psm16_Cvt;
	}
	else
	{
		m_pTexUploader_Psm16 = &CGSH_OpenGL::TexUploader_Psm16_Hw;
	}

	if(!m_nIsRGBA5551Supported || (glColorTableEXT == NULL))
	{
		m_pCvtBuffer = (uint8*)malloc(CVTBUFFERSIZE);
	}

	SetViewport(512, 384);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	SwapBuffers(m_hDC);
}

void CGSH_OpenGL::VerifyRGBA5551Support()
{
	unsigned int nTexture;

	m_nIsRGBA5551Supported = true;

	glGenTextures(1, &nTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1, 32, 32, 0, GL_RGBA, GL_UNSIGNED_SHORT_1_5_5_5_REV, m_pRAM);
	if(glGetError() == GL_INVALID_ENUM)
	{
		m_nIsRGBA5551Supported = false;
	}
	glDeleteTextures(1, &nTexture);
}

void CGSH_OpenGL::LinearZOrtho(double nLeft, double nRight, double nBottom, double nTop)
{
	double nMatrix[16];

	nMatrix[ 0] = 2.0 / (nRight - nLeft);
	nMatrix[ 1] = 0;
	nMatrix[ 2] = 0;
	nMatrix[ 3] = 0;

	nMatrix[ 4] = 0;
	nMatrix[ 5] = 2.0 / (nTop - nBottom);
	nMatrix[ 6] = 0;
	nMatrix[ 7] = 0;

	nMatrix[ 8] = 0;
	nMatrix[ 9] = 0;
	nMatrix[10] = 1;
	nMatrix[11] = 0;

	nMatrix[12] = - (nRight + nLeft) / (nRight - nLeft);
	nMatrix[13] = - (nTop + nBottom) / (nTop - nBottom);
	nMatrix[14] = 0;
	nMatrix[15] = 1;

	glMultMatrixd(nMatrix);
}

void CGSH_OpenGL::UpdateViewport()
{
	GSDISPLAY d;
	unsigned int nW, nH;

	if(m_nPMODE == 0) return;

	if(m_nPMODE & 0x01)
	{
		DECODE_DISPLAY(m_nDISPLAY1, d);
	}
	else
	{
		DECODE_DISPLAY(m_nDISPLAY2, d);
	}

	nW = (d.nW + 1) / (d.nMagX + 1);
	nH = (d.nH + 1);

	if(GetCrtIsInterlaced() && GetCrtIsFrameMode())
	{
		nH /= 2;
	}

	if(m_nWidth == nW && m_nHeight == nH)
	{
		return;
	}

	m_nWidth = nW;
	m_nHeight = nH;

//	SetViewport(m_nWidth, m_nHeight);
	SetViewport(GetCrtWidth(), GetCrtHeight());
//	SetViewport(320, 240);
	SetReadCircuitMatrix(m_nWidth, m_nHeight);

	/*
	float nRatio;
	nRatio = (float)GetCrtWidth() / (float)m_nWidth;

	if((nRatio - (int)nRatio) != 0)
	{
		glEnable(GL_LINE_SMOOTH);
	}
	else
	{
		glDisable(GL_LINE_SMOOTH);
	}

	glLineWidth(nRatio);
	*/
}

void CGSH_OpenGL::SetViewport(int nWidth, int nHeight)
{
	RECT rc;

	SetRect(&rc, 0, 0, nWidth, nHeight);
	AdjustWindowRect(&rc, GetWindowLong(m_pOutputWnd->m_hWnd, GWL_STYLE), FALSE);
	m_pOutputWnd->SetSize((rc.right - rc.left), (rc.bottom - rc.top));

	glViewport(0, 0, nWidth, nHeight);
}

void CGSH_OpenGL::SetReadCircuitMatrix(int nWidth, int nHeight)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
//	glOrtho(0, nWidth, nHeight, 0, 0, -1.0);
//	glOrtho(0, nWidth, nHeight, 0, -1.0, -65536.0);
//	glOrtho(0, nWidth, nHeight, 0, 0.01, 1000);

	LinearZOrtho(0, nWidth, nHeight, 0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

unsigned int CGSH_OpenGL::LoadTexture(GSTEX0* pReg0, GSTEX1* pReg1)
{
	unsigned int nTexture;
	int nMagFilter, nMinFilter;
	uint32 nPointer, nWidth, nHeight;
	GSTEXA TexA;

	nPointer	= pReg0->GetBufPtr();
	nWidth		= pReg0->GetWidth();
	nHeight		= pReg0->GetHeight();

	//Set the right texture function
	switch(pReg0->nFunction)
	{
	case 0:
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		break;
	case 1:
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		break;
	default:
		assert(0);
		break;
	}

	nTexture = TexCache_Search(nPointer);
	if(nTexture != 0) return nTexture;

	DECODE_TEXA(m_nReg[GS_REG_TEXA], TexA);

	glGenTextures(1, &nTexture);

	glBindTexture(GL_TEXTURE_2D, nTexture);

	//Fix the pitch of the image
	glPixelStorei(GL_UNPACK_ROW_LENGTH, pReg0->GetBufWidth());

	switch(pReg0->nPsm)
	{
	case 0:
		glPixelTransferf(GL_ALPHA_SCALE, 2.0f);
		glTexImage2D(GL_TEXTURE_2D, 0, 4, nWidth, nHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pRAM + nPointer);
		break;
	case 2:
		((this)->*(m_pTexUploader_Psm16))(pReg0, &TexA);
		break;
	case 0x13:
		((this)->*(m_pTexUploader_Psm8))(pReg0, &TexA);
		break;
	}

	if(pReg1->nMagFilter == 0)
	{
		nMagFilter = GL_NEAREST;
	}
	else
	{
		nMagFilter = GL_LINEAR;
	}

	switch(pReg1->nMinFilter)
	{
	case 0:
		nMinFilter = GL_NEAREST;
		break;
	case 1:
		nMinFilter = GL_LINEAR;
		break;
	default:
		assert(0);
		break;
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, nMagFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, nMinFilter);

	//
	/*
	glColor4d(1.0, 1.0, 1.0, 1.0);

	glBegin(GL_QUADS);
		
		glTexCoord2d(0.0,	0.0);
		glVertex2d(0.0,		0.0);

		glTexCoord2d(0.0,	1.0);
		glVertex2d(0.0,		nHeight);

		glTexCoord2d(1.0,	1.0);
		glVertex2d(nWidth,	nHeight);

		glTexCoord2d(1.0,	0.0);
		glVertex2d(nWidth,	0.0);

	glEnd();

	Flip();
	*/
	//

	glBindTexture(GL_TEXTURE_2D, 0);

	TexCache_Insert(nPointer, nWidth * nHeight * 4, nTexture);

	return nTexture;
}

void CGSH_OpenGL::TexUploader_Psm8_Hw(GSTEX0* pReg0, GSTEXA* pTexA)
{
	unsigned int nWidth, nHeight;
	uint32 nPointer;
	uint32 nPalette[0x100];
	uint32* pCLUT;
	uint32 i, j, nPos;
	uint8 nAlpha;

	nPointer	= pReg0->GetBufPtr();
	nWidth		= pReg0->GetWidth();
	nHeight		= pReg0->GetHeight();

	//Load CLUT
	//----------------------------------------

	pCLUT = (uint32*)(m_pRAM + pReg0->GetCLUTPtr());

	for(i = 0; i < 16; i++)
	{
		nPos = ((i & 1) * 8) + ((i / 2) * 0x20);

		for(j = 0; j < 8; j++)
		{
			nPalette[nPos + j] = pCLUT[j];

			nAlpha = *((uint8*)&nPalette[nPos + j] + 3);
			nAlpha = (uint8)min((unsigned int)nAlpha * 2, 0xFF);
			*((uint8*)&nPalette[nPos + j] + 3) = nAlpha;
		}
		pCLUT += 0x08;

		for(j = 0; j < 8; j++)
		{
			nPalette[nPos + j + 0x10] = pCLUT[j];

			nAlpha = *((uint8*)&nPalette[nPos + j + 0x10] + 3);
			nAlpha = (uint8)min((unsigned int)nAlpha * 2, 0xFF);
			*((uint8*)&nPalette[nPos + j + 0x10] + 3) = nAlpha;
		}
		pCLUT += 0x08;

		pCLUT += 0x30;
	}

	glColorTableEXT(GL_TEXTURE_2D, GL_RGBA, 256, GL_RGBA, GL_UNSIGNED_BYTE, nPalette);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_COLOR_INDEX8_EXT, nWidth, nHeight, 0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, m_pRAM + nPointer);
}

void CGSH_OpenGL::TexUploader_Psm8_Cvt(GSTEX0* pReg0, GSTEXA* pTexA)
{
	unsigned int nWidth, nHeight, nSrcPitch, nDstPitch, i, j, nLine;
	uint32 nPointer;
	uint32* pDst;
	uint32* pCLUT;
	uint8* pSrc;
	uint8 nColor;

	nPointer	= pReg0->GetBufPtr();
	nWidth		= pReg0->GetWidth();
	nHeight		= pReg0->GetHeight();

	nSrcPitch	= pReg0->GetBufWidth();
	nDstPitch	= nWidth;

	pSrc	= m_pRAM + nPointer;
	pDst	= (uint32*)m_pCvtBuffer;
	pCLUT	= (uint32*)(m_pRAM + pReg0->GetCLUTPtr());

	for(j = 0; j < nHeight; j++)
	{
		for(i = 0; i < nWidth; i++)
		{
			nColor = pSrc[i];
			//Fix index for CLUT arrangement
			nColor = (nColor & ~0x18) | ((nColor & 0x08) << 1) | ((nColor & 0x10) >> 1);
			nLine = (nColor & 0xF0) >> 4;
			pDst[i] = pCLUT[(nColor & 0x0F) + (nLine * 0x40)];
		}
		
		pSrc += nSrcPitch;
		pDst += nDstPitch;
	}

	glPixelTransferf(GL_ALPHA_SCALE, 2.0f);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, nWidth);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, nWidth, nHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pCvtBuffer);
}

void CGSH_OpenGL::TexUploader_Psm16_Hw(GSTEX0* pReg0, GSTEXA* pTexA)
{
	unsigned int nWidth, nHeight;
	uint32 nPointer;

	nPointer	= pReg0->GetBufPtr();
	nWidth		= pReg0->GetWidth();
	nHeight		= pReg0->GetHeight();

	//No clue about this... 'lobotom.elf'
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0x280);

	if(pReg0->nColorComp == 1)
	{
		//This isn't right... TA0 isn't considered
		glPixelTransferf(GL_ALPHA_SCALE, (float)pTexA->nTA1);
	}
	else
	{
		glPixelTransferf(GL_ALPHA_SCALE, 0);
	}

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1, nWidth, nHeight, 0, GL_RGBA, GL_UNSIGNED_SHORT_1_5_5_5_REV, m_pRAM + nPointer);
}

void CGSH_OpenGL::TexUploader_Psm16_Cvt(GSTEX0* pReg0, GSTEXA* pTexA)
{
	unsigned int nWidth, nHeight, nSrcPitch, nDstPitch, i, j;
	uint8 nTA0, nTA1;
	uint8 nR, nG, nB, nA;
	uint16* pSrc;
	uint32* pDst;
	uint16 nPixel;
	uint32 nPointer;

	nPointer	= pReg0->GetBufPtr();
	nWidth		= pReg0->GetWidth();
	nHeight		= pReg0->GetHeight();

	//No clue about this... 'lobotom.elf'
	nSrcPitch = 0x280;
	nDstPitch = nWidth;

	pSrc = (uint16*)(m_pRAM + nPointer);
	pDst = (uint32*)(m_pCvtBuffer);

	if(pReg0->nColorComp == 1)
	{
		nTA0 = pTexA->nTA0;
		nTA1 = pTexA->nTA1;
	}
	else
	{
		nTA0 = 0;
		nTA1 = 0;
	}

	for(j = 0; j < nHeight; j++)
	{
		for(i = 0; i < nWidth; i++)
		{
			nPixel = pSrc[i];
			
			nB = (nPixel >>  0) & 0x1F;
			nG = (nPixel >>  5) & 0x1F;
			nR = (nPixel >> 10) & 0x1F;
			nA = (nPixel >> 15) & 0x01;

			nB <<= 3;
			nG <<= 3;
			nR <<= 3;
			nA = (nA == 0) ? nTA0 : nTA1;

			*(((uint8*)&pDst[i]) + 0) = nB;
			*(((uint8*)&pDst[i]) + 1) = nG;
			*(((uint8*)&pDst[i]) + 2) = nR;
			*(((uint8*)&pDst[i]) + 3) = nA;
		}

		pSrc += nSrcPitch;
		pDst += nDstPitch;
	}

	glPixelTransferf(GL_ALPHA_SCALE, 2.0f);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, nWidth);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, nWidth, nHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pCvtBuffer);
}

double CGSH_OpenGL::GetZ(double nZ)
{
	if(nZ == 0)
	{
		return -1;
	}
	if(nZ < 256)
	{
		//The number is small, so scale to a smaller ratio (65536)
		return (nZ - 32768.0) / 32768.0;
	}
	else
	{
		return (nZ - m_nMaxZ) / m_nMaxZ;
	}
}

unsigned int CGSH_OpenGL::GetNextPowerOf2(unsigned int nNumber)
{
	return 1 << ((int)(log((double)(nNumber - 1)) / log(2.0)) + 1);
}

/////////////////////////////////////////////////////////////
// Texture
/////////////////////////////////////////////////////////////

CGSH_OpenGL::CTexture::CTexture()
{
	m_nStart	= 0;
	m_nSize		= 0;
	m_nTexture	= 0;
}

CGSH_OpenGL::CTexture::~CTexture()
{
	Invalidate();
}

void CGSH_OpenGL::CTexture::Invalidate()
{
	if(m_nTexture != 0)
	{
		glDeleteTextures(1, &m_nTexture);
		m_nTexture = 0;
	}
}

void CGSH_OpenGL::CTexture::InvalidateFromMemorySpace(uint32 nStart, uint32 nSize)
{
	bool nInvalid;
	uint32 nTexEnd, nEnd;

	if(!IsValid()) return;
	
	nInvalid = false;
	nEnd = nStart + nSize;
	nTexEnd = m_nStart + m_nSize;

	if((nStart >= m_nStart) && (nStart <= nTexEnd))
	{
		nInvalid = true;
	}

	if((nEnd >= m_nStart) && (nEnd <= nTexEnd))
	{
		nInvalid = true;
	}

	if(nInvalid)
	{
		Invalidate();
	}
}

bool CGSH_OpenGL::CTexture::IsValid()
{
	return (m_nTexture != 0);
}

/////////////////////////////////////////////////////////////
// Texture Caching
/////////////////////////////////////////////////////////////

unsigned int CGSH_OpenGL::TexCache_Search(uint32 nStart)
{
	unsigned int i;

	for(i = 0; i < MAXCACHE; i++)
	{
		if(!m_TexCache[i].IsValid()) continue;
		if(m_TexCache[i].m_nStart == nStart)
		{
			return m_TexCache[i].m_nTexture;
		}
	}

	return 0;
}

void CGSH_OpenGL::TexCache_Insert(uint32 nStart, uint32 nSize, unsigned int nTexture)
{
	CTexture* pTexture;
	pTexture = &m_TexCache[m_nTexCacheIndex];

	pTexture->Invalidate();

	pTexture->m_nStart		= nStart;
	pTexture->m_nSize		= nSize;
	pTexture->m_nTexture	= nTexture;

	m_nTexCacheIndex++;
	m_nTexCacheIndex %= MAXCACHE;
}

void CGSH_OpenGL::TexCache_InvalidateTextures(uint32 nStart, uint32 nSize)
{
	unsigned int i;

	for(i = 0; i < MAXCACHE; i++)
	{
		m_TexCache[i].InvalidateFromMemorySpace(nStart, nSize);
	}
}

/////////////////////////////////////////////////////////////
// Context Unpacking
/////////////////////////////////////////////////////////////

void CGSH_OpenGL::SetRenderingContext(unsigned int nContext)
{
	SetupBlendingFunction(m_nReg[GS_REG_ALPHA_1 + nContext]);
	SetupTestFunctions(m_nReg[GS_REG_TEST_1 + nContext]);
	SetupDepthBuffer(m_nReg[GS_REG_ZBUF_1 + nContext]);
	SetupTexture(m_nReg[GS_REG_TEX0_1 + nContext], m_nReg[GS_REG_TEX1_1 + nContext]);
	
	DECODE_XYOFFSET(m_nReg[GS_REG_XYOFFSET_1 + nContext], m_nPrimOfsX, m_nPrimOfsY);
	
	if(GetCrtIsInterlaced() && GetCrtIsFrameMode())
	{
		if(m_nCSR & 0x2000)
		{
			m_nPrimOfsY += 0.5;
		}
	}
}

void CGSH_OpenGL::SetupBlendingFunction(uint64 nData)
{
	GSALPHA alpha;
	int nFunction;

	if(nData == 0) return;

	nFunction = GL_FUNC_ADD_EXT;

	DECODE_ALPHA(nData, alpha);

	if((alpha.nA == 0) && (alpha.nB == 1) && (alpha.nC == 0) && (alpha.nD == 1))
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else if((alpha.nA == 0) && (alpha.nB == 2) && (alpha.nC == 2) && (alpha.nD == 1) && (alpha.nFix == 0x80))
	{
		glBlendFunc(GL_ONE, GL_ONE);
	}
	else if((alpha.nA == 0) && (alpha.nB == 1) && (alpha.nC == 2) && (alpha.nD == 1) && (alpha.nFix == 0x80))
	{
		glBlendFunc(GL_ONE, GL_ZERO);
	}
	else if((alpha.nA == 0) && (alpha.nB == 1) && (alpha.nC == 2) && (alpha.nD == 1))
	{
		//Source alpha value is implied in the formula
		//As = FIX / 0x80
		if(glBlendColorEXT != NULL)
		{
			glBlendColorEXT(0.0f, 0.0f, 0.0f, (float)alpha.nFix / 128.0f);
			glBlendFunc(GL_CONSTANT_ALPHA_EXT, GL_ONE_MINUS_CONSTANT_ALPHA_EXT);
		}
	}
	else if((alpha.nA == 1) && (alpha.nB == 0) && (alpha.nC == 2) && (alpha.nD == 2))
	{
		nFunction = GL_FUNC_REVERSE_SUBTRACT_EXT;
		if(glBlendColorEXT != NULL)
		{
			glBlendColorEXT(0.0f, 0.0f, 0.0f, (float)alpha.nFix / 128.0f);
			glBlendFunc(GL_CONSTANT_ALPHA_EXT, GL_CONSTANT_ALPHA_EXT);
		}
	}
	else if((alpha.nA == 0) && (alpha.nB == 2) && (alpha.nC == 0) && (alpha.nD == 1))
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	}
	else
	{
		printf("GSH_OpenGL: Unknown color blending formula.\r\n");
	}

	if(glBlendEquationEXT != NULL)
	{
		glBlendEquationEXT(nFunction);
	}
}

void CGSH_OpenGL::SetupTestFunctions(uint64 nData)
{
	GSTEST tst;
	unsigned int nFunc;
	float nValue;

	DECODE_TEST(nData, tst);

	if(tst.nAlphaEnabled)
	{
		switch(tst.nAlphaMethod)
		{
		case 0:
			nFunc = GL_NEVER;
			break;
		case 1:
			nFunc = GL_ALWAYS;
			break;
		case 2:
			nFunc = GL_LESS;
			break;
		case 7:
			nFunc = GL_NOTEQUAL;
			break;
		default:
			assert(0);
			break;
		}

		nValue = (float)tst.nAlphaRef / 255.0f;
		glAlphaFunc(nFunc, nValue);

		glEnable(GL_ALPHA_TEST);

	}
	else
	{
		glDisable(GL_ALPHA_TEST);
	}

	if(tst.nDepthEnabled)
	{
		switch(tst.nDepthMethod)
		{
		case 0:
			nFunc = GL_NEVER;
			break;
		case 1:
			nFunc = GL_ALWAYS;
			break;
		case 2:
			nFunc = GL_GEQUAL;
			break;
		case 3:
			nFunc = GL_GREATER;
			break;
		}

		glDepthFunc(nFunc);

		glEnable(GL_DEPTH_TEST);
	}
	else
	{
		glDisable(GL_DEPTH_TEST);
	}
}

void CGSH_OpenGL::SetupDepthBuffer(uint64 nData)
{
	GSZBUF zbuf;

	DECODE_ZBUF(nData, zbuf);

	switch(GetPsmPixelSize(zbuf.nPsm))
	{
	case 2:
		m_nMaxZ = 32768.0;
		break;
	case 4:
		m_nMaxZ = 2147483647.0;
		break;
	}
}

void CGSH_OpenGL::SetupTexture(uint64 nTex0, uint64 nTex1)
{
	GSTEX0 tex0;
	GSTEX1 tex1;

	if(nTex0 == 0)
	{
		m_nTexHandle = 0;
		return;
	}

	DECODE_TEX0(nTex0, tex0);
	DECODE_TEX1(nTex1, tex1);

	m_nTexWidth		= tex0.GetWidth();
	m_nTexHeight	= tex0.GetHeight();
	m_nTexHandle	= LoadTexture(&tex0, &tex1);
}

/////////////////////////////////////////////////////////////
// Individual Primitives Implementations
/////////////////////////////////////////////////////////////

void CGSH_OpenGL::Prim_Point()
{
	double nX, nY, nZ;
	GSRGBAQ rgbaq;
	
	DECODE_XYZ2(m_VtxBuffer[0].nPosition, nX, nY, nZ);

	nX -= m_nPrimOfsX;
	nY -= m_nPrimOfsY;

	if(!m_PrimitiveMode.nUseUV && !m_PrimitiveMode.nTexture)
	{
		DECODE_RGBAQ(m_VtxBuffer[0].nRGBAQ, rgbaq);

		glBegin(GL_QUADS);

			glColor4ub(rgbaq.nR, rgbaq.nG, rgbaq.nB, rgbaq.nA);

			glVertex2d(nX + 0, nY + 0);
			glVertex2d(nX + 1, nY + 0);
			glVertex2d(nX + 1, nY + 1);
			glVertex2d(nX + 0, nY + 1);

		glEnd();

	}
	else
	{
		//Yay for textured points!
		assert(0);
	}
}

void CGSH_OpenGL::Prim_Line()
{
	double nX1, nX2;
	double nY1, nY2;
	double nZ;
	GSRGBAQ rgbaq[2];
	
	DECODE_XYZ2(m_VtxBuffer[1].nPosition, nX1, nY1, nZ);
	DECODE_XYZ2(m_VtxBuffer[0].nPosition, nX2, nY2, nZ);

	nX1 -= m_nPrimOfsX;
	nX2 -= m_nPrimOfsX;

	nY1 -= m_nPrimOfsY;
	nY2 -= m_nPrimOfsY;

	if(m_PrimitiveMode.nAntiAliasing)
	{
		glEnable(GL_BLEND);
	}

	if(!m_PrimitiveMode.nUseUV && !m_PrimitiveMode.nTexture)
	{

		DECODE_RGBAQ(m_VtxBuffer[1].nRGBAQ, rgbaq[0]);
		DECODE_RGBAQ(m_VtxBuffer[0].nRGBAQ, rgbaq[1]);

		if(m_nLinesAsQuads)
		{
			glBegin(GL_QUADS);

				glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, rgbaq[0].nA);
		
				glVertex2d(nX1 + 0, nY1 + 0);
				glVertex2d(nX1 + 1, nY1 + 1);

				glColor4ub(rgbaq[1].nR, rgbaq[1].nG, rgbaq[1].nB, rgbaq[1].nA);

				glVertex2d(nX2 + 1, nY2 + 1);
				glVertex2d(nX2 + 0, nY2 + 0);

			glEnd();
		}
		else
		{
			glBegin(GL_LINES);

				glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, rgbaq[0].nA);
				glVertex2d(nX1, nY1);

				glColor4ub(rgbaq[1].nR, rgbaq[1].nG, rgbaq[1].nB, rgbaq[1].nA);
				glVertex2d(nX2, nY2);

			glEnd();
		}
	}
	else
	{

		//Yay for textured lines!
		assert(0);
	}

	if(m_PrimitiveMode.nAntiAliasing)
	{
		glDisable(GL_BLEND);
	}
}

void CGSH_OpenGL::Prim_Triangle()
{
	double nX1, nX2, nX3;
	double nY1, nY2, nY3;
	double nZ1, nZ2, nZ3;

	double nU1, nU2, nU3;
	double nV1, nV2, nV3;

	double nS1, nS2, nS3;
	double nT1, nT2, nT3;

	GSRGBAQ rgbaq[3];

	DECODE_XYZ2(m_VtxBuffer[2].nPosition, nX1, nY1, nZ1);
	DECODE_XYZ2(m_VtxBuffer[1].nPosition, nX2, nY2, nZ2);
	DECODE_XYZ2(m_VtxBuffer[0].nPosition, nX3, nY3, nZ3);

	DECODE_RGBAQ(m_VtxBuffer[2].nRGBAQ, rgbaq[0]);
	DECODE_RGBAQ(m_VtxBuffer[1].nRGBAQ, rgbaq[1]);
	DECODE_RGBAQ(m_VtxBuffer[0].nRGBAQ, rgbaq[2]);

	nX1 -= m_nPrimOfsX;
	nX2 -= m_nPrimOfsX;
	nX3 -= m_nPrimOfsX;

	nY1 -= m_nPrimOfsY;
	nY2 -= m_nPrimOfsY;
	nY3 -= m_nPrimOfsY;

	nZ1 = GetZ(nZ1);
	nZ2 = GetZ(nZ2);
	nZ3 = GetZ(nZ3);

	if(m_PrimitiveMode.nShading)
	{
		glShadeModel(GL_SMOOTH);
	}
	else
	{
		glShadeModel(GL_FLAT);
	}

	if(m_PrimitiveMode.nAlpha)
	{
		glEnable(GL_BLEND);
	}

	if(m_PrimitiveMode.nTexture)
	{
		//Textured triangle

		if(m_PrimitiveMode.nUseUV)
		{
			glColor4d(1.0, 1.0, 1.0, 1.0);

			glBindTexture(GL_TEXTURE_2D, m_nTexHandle);

			DECODE_UV(m_VtxBuffer[2].nUV, nU1, nV1);
			DECODE_UV(m_VtxBuffer[1].nUV, nU2, nV2);
			DECODE_UV(m_VtxBuffer[0].nUV, nU3, nV3);

			nU1 /= (double)m_nTexWidth;
			nU2 /= (double)m_nTexWidth;
			nU3 /= (double)m_nTexWidth;

			nV1 /= (double)m_nTexHeight;
			nV2 /= (double)m_nTexHeight;
			nV3 /= (double)m_nTexHeight;

			glBegin(GL_TRIANGLES);
				
				glTexCoord2d(nU1, nV1);
				glVertex2d(nX1, nY1);

				glTexCoord2d(nU2, nV2);
				glVertex2d(nX2, nY2);

				glTexCoord2d(nU3, nV3);
				glVertex2d(nX3, nY3);

			glEnd();

			glBindTexture(GL_TEXTURE_2D, NULL);

		}
		else
		{
			glBindTexture(GL_TEXTURE_2D, m_nTexHandle);

			DECODE_ST(m_VtxBuffer[2].nST, nS1, nT1);
			DECODE_ST(m_VtxBuffer[1].nST, nS2, nT2);
			DECODE_ST(m_VtxBuffer[0].nST, nS3, nT3);

			nS1 /= rgbaq[0].nQ;
			nS2 /= rgbaq[1].nQ;
			nS3 /= rgbaq[2].nQ;

			nT1 /= rgbaq[0].nQ;
			nT2 /= rgbaq[1].nQ;
			nT3 /= rgbaq[2].nQ;

			glBegin(GL_TRIANGLES);
				
				glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, rgbaq[0].nA * 2);
				glTexCoord2d(nS1, nT1);
				glVertex3d(nX1, nY1, nZ1);

				glColor4ub(rgbaq[1].nR, rgbaq[1].nG, rgbaq[1].nB, rgbaq[1].nA * 2);
				glTexCoord2d(nS2, nT2);
				glVertex3d(nX2, nY2, nZ2);

				glColor4ub(rgbaq[2].nR, rgbaq[2].nG, rgbaq[2].nB, rgbaq[2].nA * 2);
				glTexCoord2d(nS3, nT3);
				glVertex3d(nX3, nY3, nZ3);

			glEnd();

			glBindTexture(GL_TEXTURE_2D, NULL);			
		}
	}
	else
	{
		//Non Textured Triangle
		glBegin(GL_TRIANGLES);
			
			glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, (uint8)min(0xFF, (int)rgbaq[0].nA * 2));
			glVertex3d(nX1, nY1, nZ1);

			glColor4ub(rgbaq[1].nR, rgbaq[1].nG, rgbaq[1].nB, (uint8)min(0xFF, (int)rgbaq[1].nA * 2));
			glVertex3d(nX2, nY2, nZ2);

			glColor4ub(rgbaq[2].nR, rgbaq[2].nG, rgbaq[2].nB, (uint8)min(0xFF, (int)rgbaq[2].nA * 2));
			glVertex3d(nX3, nY3, nZ3);

		glEnd();
	}

	if(m_PrimitiveMode.nAlpha)
	{
		glDisable(GL_BLEND);
	}
}

void CGSH_OpenGL::Prim_Sprite()
{
	double nX1, nX2;
	double nY1, nY2;
	double nZ1, nZ2;

	double nU1, nU2;
	double nV1, nV2;

	GSRGBAQ rgbaq[2];

	DECODE_XYZ2(m_VtxBuffer[1].nPosition, nX1, nY1, nZ1);
	DECODE_XYZ2(m_VtxBuffer[0].nPosition, nX2, nY2, nZ2);

	DECODE_RGBAQ(m_VtxBuffer[1].nRGBAQ, rgbaq[0]);
	DECODE_RGBAQ(m_VtxBuffer[0].nRGBAQ, rgbaq[1]);

	nX1 -= m_nPrimOfsX;
	nX2 -= m_nPrimOfsX;

	nY1 -= m_nPrimOfsY;
	nY2 -= m_nPrimOfsY;

//	assert(nZ1 == nZ2);

	nZ1 = GetZ(nZ1);
	nZ2 = GetZ(nZ2);

	if(m_PrimitiveMode.nAlpha)
	{
		glEnable(GL_BLEND);
	}

	if(m_PrimitiveMode.nUseUV && m_PrimitiveMode.nTexture)
	{
		glColor4d(1.0, 1.0, 1.0, 1.0);
//		glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, (uint8)min(0xFF, (int)rgbaq[0].nA * 2));

		glBindTexture(GL_TEXTURE_2D, m_nTexHandle);

		DECODE_UV(m_VtxBuffer[1].nUV, nU1, nV1);
		DECODE_UV(m_VtxBuffer[0].nUV, nU2, nV2);

		nU1 /= (double)m_nTexWidth;
		nU2 /= (double)m_nTexWidth;

		nV1 /= (double)m_nTexHeight;
		nV2 /= (double)m_nTexHeight;

		glBegin(GL_QUADS);

			glTexCoord2d(nU1, nV1);
			glVertex3d(nX1, nY1, nZ1);

			glTexCoord2d(nU2, nV1);
			glVertex3d(nX2, nY1, nZ2);

			glTexCoord2d(nU2, nV2);
			glVertex3d(nX2, nY2, nZ1);

			glTexCoord2d(nU1, nV2);
			glVertex3d(nX1, nY2, nZ2);

		glEnd();

		glBindTexture(GL_TEXTURE_2D, 0);
	}
	else if(!m_PrimitiveMode.nUseUV && !m_PrimitiveMode.nTexture)
	{
		//Humm? Would it be possible to have a gradient using those registers?

		glColor4ub(rgbaq[0].nR, rgbaq[0].nG, rgbaq[0].nB, rgbaq[0].nA);

		glBegin(GL_QUADS);

			glVertex3d(nX1, nY1, nZ1);
			glVertex3d(nX2, nY1, nZ2);
			glVertex3d(nX2, nY2, nZ1);
			glVertex3d(nX1, nY2, nZ2);

		glEnd();

	}
	else
	{
		assert(0);
	}

	if(m_PrimitiveMode.nAlpha)
	{
		glDisable(GL_BLEND);
	}
}

/////////////////////////////////////////////////////////////
// Other Functions
/////////////////////////////////////////////////////////////

void CGSH_OpenGL::WriteRegister(uint8 nRegister, uint64 nData)
{
	CGSHandler::WriteRegister(nRegister, nData);
	
	switch(nRegister)
	{
	case GS_REG_PRIM:
		m_nPrimitiveType = (unsigned int)(nData & 0x07);
		switch(m_nPrimitiveType)
		{
		case 0:
			//Point
			m_nVtxCount = 1;
			break;
		case 1:
			//Line
			m_nVtxCount = 2;
			break;
		case 2:
			//Line strip
			m_nVtxCount = 2;
			break;
		case 3:
			//Triangle
			m_nVtxCount = 3;
			break;
		case 4:
			//Triangle Strip
			m_nVtxCount = 3;
			break;
		case 6:
			//Sprite (rectangle)
			m_nVtxCount = 2;
			break;
		default:
			printf("GS: Unhandled primitive type (%i) encountered.\r\n", m_nPrimitiveType);
			break;
		}
		break;

	case GS_REG_XYZ2:
	case GS_REG_XYZ3:
		if(m_nVtxCount == 0) break;

		m_VtxBuffer[m_nVtxCount - 1].nPosition	= nData;
		m_VtxBuffer[m_nVtxCount - 1].nRGBAQ		= m_nReg[GS_REG_RGBAQ];
		m_VtxBuffer[m_nVtxCount - 1].nUV		= m_nReg[GS_REG_UV];
		m_VtxBuffer[m_nVtxCount - 1].nST		= m_nReg[GS_REG_ST];

		m_nVtxCount--;

		if(m_nVtxCount == 0)
		{
			if((m_nReg[GS_REG_PRMODECONT] & 1) != 0)
			{
				DECODE_PRMODE(m_nReg[GS_REG_PRIM], m_PrimitiveMode);
			}
			else
			{
				DECODE_PRMODE(m_nReg[GS_REG_PRMODE], m_PrimitiveMode);
			}

			SetRenderingContext(m_PrimitiveMode.nContext);

			switch(m_nPrimitiveType)
			{
			case 0:
				if(nRegister == GS_REG_XYZ2) Prim_Point();
				break;
			case 1:
				if(nRegister == GS_REG_XYZ2) Prim_Line();
				break;
			case 2:
				if(nRegister == GS_REG_XYZ2) Prim_Line();
				memcpy(&m_VtxBuffer[1], &m_VtxBuffer[0], sizeof(VERTEX));
				m_nVtxCount = 1;
				break;
			case 3:
				if(nRegister == GS_REG_XYZ2) Prim_Triangle();
				m_nVtxCount = 3;
				break;
			case 4:
				if(nRegister == GS_REG_XYZ2) Prim_Triangle();
				memcpy(&m_VtxBuffer[2], &m_VtxBuffer[1], sizeof(VERTEX));
				memcpy(&m_VtxBuffer[1], &m_VtxBuffer[0], sizeof(VERTEX));
				m_nVtxCount = 1;
				break;
			case 6:
				if(nRegister == GS_REG_XYZ2) Prim_Sprite();
				m_nVtxCount = 2;
				break;
			}
		}
		break;
	}
}

void CGSH_OpenGL::ProcessImageTransfer(uint32 nAddress, uint32 nLenght)
{
	GSBITBLTBUF buf;
	GSFRAME fr;
	uint32 nFrameEnd;

	DECODE_BITBLTBUF(m_nReg[GS_REG_BITBLTBUF], buf);
	DECODE_FRAME(m_nReg[GS_REG_FRAME_1], fr);

	nFrameEnd = (fr.GetBasePtr() + (fr.GetWidth() * GetPsmPixelSize(fr.nPsm)) * m_nHeight);
	if(nAddress >= nFrameEnd)
	{
		//Probably just a texture upload...
		TexCache_InvalidateTextures(nAddress, nLenght);
	}
	else
	{
		DisplayTransferedImage(nAddress);
	}
}

void CGSH_OpenGL::DisplayTransferedImage(uint32 nAddress)
{
	unsigned int nW, nH;
	unsigned int nW2, nH2;
	unsigned int nSX, nSY, nDX, nDY, nD;
	unsigned int nTexture;
	uint32* pImg;
	double nS, nT;

	DECODE_TRXREG(m_nReg[GS_REG_TRXREG], nW, nH);
	DECODE_TRXPOS(m_nReg[GS_REG_TRXPOS], nSX, nSY, nDX, nDY, nD);	

	pImg = (uint32*)(m_pRAM + nAddress);

	glGenTextures(1, &nTexture);

	glBindTexture(GL_TEXTURE_2D, nTexture);

	nW2 = GetNextPowerOf2(nW);
	nH2 = GetNextPowerOf2(nH);

	//Upload the texture
	glPixelStorei(GL_UNPACK_ROW_LENGTH, nW);

	glTexImage2D(GL_TEXTURE_2D, 0, 4, nW2, nH2, 0, GL_RGBA, GL_UNSIGNED_BYTE, pImg);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	glColor4d(1.0, 1.0, 1.0, 1.0);

	nS = (double)nW / (double)nW2;
	nT = (double)nH / (double)nH2;

	glBegin(GL_QUADS);
		
		glTexCoord2d(0.0,		0.0);
		glVertex2d(nDX,			nDY);

		glTexCoord2d(0.0,		nT);
		glVertex2d(nDX,			nDY + nH);

		glTexCoord2d(nS,		nT);
		glVertex2d(nDX + nW,	nDY + nH);

		glTexCoord2d(nS,		0.0);
		glVertex2d(nDX + nW,	nDY);

	glEnd();

	glBindTexture(GL_TEXTURE_2D, NULL);

	glDeleteTextures(1, &nTexture);
}

void CGSH_OpenGL::Flip()
{
	SwapBuffers(m_hDC);
}

CSettingsDialogProvider* CGSH_OpenGL::GetSettingsDialogProvider()
{
	return this;
}

CModalWindow* CGSH_OpenGL::CreateSettingsDialog(HWND hParent)
{
	return new CRendererSettingsWnd(hParent, this);
}

void CGSH_OpenGL::OnSettingsDialogDestroyed()
{
	m_nLinesAsQuads = CConfig::GetInstance()->GetPreferenceBoolean(PREF_CGSH_OPENGL_LINEASQUADS);
}

bool CGSH_OpenGL::IsColorTableExtSupported()
{
	return glColorTableEXT != NULL;
}

bool CGSH_OpenGL::IsBlendColorExtSupported()
{
	return glBlendColorEXT != NULL;
}

bool CGSH_OpenGL::IsBlendEquationExtSupported()
{
	return glBlendEquationEXT != NULL;
}

bool CGSH_OpenGL::IsRGBA5551ExtSupported()
{
	return m_nIsRGBA5551Supported;
}

CGSHandler* CGSH_OpenGL::GSHandlerFactory(void* pParam)
{
	return new CGSH_OpenGL((CWindow*)pParam);
}
